/* src/serveur.c - Fonctions serveur
 * Copyright (C) 2002-2003 Inter System
 *
 * contact: Progs@Inter-System.Net
 *          Cesar@Inter-System.Net
 *          kouak@kouak.org
 * site web: http://coderz.inter-system.net
 *
 * Services pour serveur IRC. Support sur IrcProgs
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <ctype.h>
#include <unistd.h>
#include <arpa/inet.h>
#include <sys/time.h>
#include <time.h>
#include <sys/socket.h>
#include "main.h"

CVSINFO("$Id: serveur.c,v 1.7 2003/02/16 19:37:47 romexzf Exp $");

struct Link *mainhub;
int dataQ = 0;
int burst;
int registered;
int want_to_quit;
extern const int STS;
extern const int kill_interval;
extern const char pasdeperm[];
extern const char welcome[];
extern const int ConfFlag;

extern int lastresetparse;
extern int lastresetlogs;

extern struct cs_commands cs_cmds[];

struct aMsg msgtab[] = {
  {TOKEN_PRIVMSG, m_privmsg},
  {TOKEN_PING, m_ping},
  {TOKEN_NICK, m_nick},
  {TOKEN_JOIN, m_join},
  {TOKEN_PART, m_part},
  {TOKEN_MODE, m_mode},
  {TOKEN_KICK, m_kick},
  {TOKEN_CREATE, m_create},
  {TOKEN_QUIT, m_quit},
  {TOKEN_ACCOUNT, m_account},
  {TOKEN_BURST, m_burst},
  {TOKEN_AWAY, m_away},
  {TOKEN_SERVER, m_server},
#ifdef N2K
  {"SERVER", m_server},
#endif
  {TOKEN_SQUIT, m_squit},
  {TOKEN_TOPIC, m_topic},
  {TOKEN_KILL, m_kill},
  {TOKEN_WHOIS, m_whois},
  {TOKEN_SETHOST, m_sethost},
  {TOKEN_SETIDENT, m_setident},
  {TOKEN_PASS, m_pass},
  {TOKEN_EOB, m_eob},
  {NULL, NULL}
};

/*
 * m_account
 *
 * parv[0] = sender (i don't care about)
 * parv[1] = numeric
 * parv[2] = login
 */
void m_account(int sock, int parc, char **parv)
{
	aNick *nick;
	anUser *user;
	char *login = NULL;

	if(parc != 3) return;
	if(!(nick=num2nickinfo(parv[1]))) return;

	login = parv[2];
	if(*login == ':') login++;

	if(!(user=getuserinfo(login)))
	{
		struct Link *link;
		link=num2servinfo(parv[0]);
		cswallops("ACCOUNT inconu (%s) pour %s depuis %s", login, nick->nick, link->serv);
		cs_account(nick, NULL); // desacounter :)
		return;
	}
	if(user->n)
	{
		struct Link *link;
		link=num2servinfo(parv[0]);
		cswallops("ACCOUNT DESYNCH (%s) pour %s depuis %s", login, nick->nick, link->serv);
		cs_account(nick, NULL); // desacounter :)
		return;
	}

	nick->user = user;
	nick->user->n = nick;
	nick->user->lastseend = time(NULL);
	csntc(nick, "Vous avez t \002ridentifi\002");
	if(UProtect(user)) del_killinfo(nick->nick, nick->numeric); /* arret du kill */
	return;
}

/*
 * m_mode
 *
 * parv[0] = who
 * parv[1] = chan
 * parv[2] = modes
 * parv[3->parc-1] = rest
 */
void m_mode(int sock, int parc, char **parv)
{
	aLink *j;
	anAccess *a = NULL, *a2 = NULL;
	aNick *nick, *nicktemp = NULL;
	aChan *ch = NULL;
	aJoin *join;
	char *modes = NULL, *param = NULL, *key = NULL, *limit = NULL;
	int what = 1, i = 2, l = 0;

	if(parc < 2) return;
	if(parv[1][0] == '#')
	{
		ch=getchaninfo(parv[1]);

		modes = parv[2];
		while (modes && *modes)
		{
			switch (*modes)
			{
				case '+':
					what = 1;
					break;
				case '-':
					what = 0;
					break;
				case 'o':
					do
					{
						i++;
						param = parv[i];
					}
					while(!*param || !param);

					if(ch) /* salon reg */
					{
						if(!what) /* -o */
						{
							if(!strcmp(cs.num, param)) /* deop du CServ */
							{
								csmode(ch, "+o %s", cs.num);
								break;
							}
							/* protection du deop des admins /user auths */
							if((nicktemp = num2nickinfo(param)) && nicktemp->user &&
#ifdef N2K
							(nick=num2nickinfo(parv[0]))
#else
							(nick=getnickbynick(parv[0]))
#endif
							&& nick != nicktemp && (IsAnAdmin(nicktemp->user) || (a=GetAccessIbyUserI(nicktemp->user, ch->chan))))
							{
								if(!nick->user || (nicktemp->user->level > nick->user->level ||
								(a && AProtect(a) && !ASuspend(a) && (a2=GetAccessIbyUserI(nick->user, ch->chan)) && !ASuspend(a2) && a2->level <= a->level)))
								{
									csmode(ch, "+o %s", nicktemp->numeric);
									break;
								}
							}
						}
						else /* +o */
						{
							nicktemp = num2nickinfo(param);
							if(CNoOps(ch) && nicktemp && !HasMode(nicktemp, 'o'))
							{
								csmode(ch, "-o %s", param);
								break;
							}
							if(CStrictOp(ch) && nicktemp && (!nicktemp->user || (!(a=GetAccessIbyUserI(nicktemp->user, ch->chan)) || ASuspend(a))))
							{
								csmode(ch, "-o %s", param);
								break;
							}
						}
						/* mise  jour des modes */
						for(j=ch->members;j && strcasecmp(j->value.j->nick->numeric, param);j=j->next);
						if(j) j->value.j->is_op = what;

					} /* mise  jour direct dans nick->join */
					else if((join = getjoininfo((nicktemp = num2nickinfo(param)), parv[1]))) join->is_op = what;

					break;

				case 'v':
					do
					{
						i++;
						param = parv[i];
					}
					while(!*param || !param);

					j=NULL;
					join=NULL;
					if(ch)
					{
						for(j=ch->members;j;j=j->next)
						if(!strcasecmp(j->value.j->nick->numeric, param)) break;
						if(j) j->value.j->is_voice = what;
					}
					else if((join=getjoininfo((nicktemp=num2nickinfo(param)), parv[1]))) join->is_voice = what;
					break;

				case 'b':
					do  /*l on modifie 'i' pour qu'il passe au paramtre de ce mode, pour pas ke celui du suivant soit celui ci, hein cesar*/
					{
						i++;
						param = parv[i];
					}
					while(!*param || !param);
					if(!what || !ch) break;
					if(!CNoBans(ch))
					{
						char h[200];
#ifdef N2K
						if(!(nicktemp=num2nickinfo(parv[0]))) break;
#else
						if(!(nicktemp=getnickbynick(parv[0]))) break;
#endif
						l = nicktemp->user ? ChanLevelbyUserI(nicktemp->user, parv[1]) : 0;
						for(j=ch->members;j;j=j->next)
						{
							nick=j->value.j->nick;
							snprintf(h, 300, "%s!%s@%s", nick->nick, nick->ident,
#ifdef HAVE_CRYPTHOST
							nick->crypt
#else
							nick->host
#endif
							);
							if(match(param, h) || !nick->user) continue;
							for(a=nick->user->accesshead;a && strcasecmp(a->chan, ch->chan);a=a->next);
							if(a && AProtect(a)) break;
							else a=NULL;
						}
					}
					if(CNoBans(ch) || ((ch->banlevel > l || (a && AProtect(a))) && (!nicktemp->user || !IsAnAdmin(nicktemp->user))))
					{
						csmode(ch, "-b %s", param);
						break;
					}
					break;
				case 'l':
					if(what)
					{
						i++;
						if(is_num(parv[i])) limit = parv[i];
					}
					break;
				case 'k':
					i++;
					key = parv[i];
					break;
			}
			modes++;
		}
		if(ch) parse_cmode(ch, parv[2], limit, key);
	}
	else if((nick = getnickbynick(parv[1]))) parse_umode(nick, parv[2]);
	return;
}

/*
 * m_part
 *
 * parv[0] = nick
 * parv[1] = chans
 * parv[2] = raison
 */
void m_part(int sock, int parc, char **parv)
{
  char *p;
  aNick *nick=NULL;
  int i = 1;

  if(parc < 2) return;
#ifdef N2K
  nick = num2nickinfo(parv[0]);
#else
  nick = getnickbynick(parv[0]);
#endif
  p = strtok(parv[1], ",");
  while(p != NULL)
  {
	  del_join(nick, p);
	  i++;
	  p = strtok(NULL, ",");
  }

  return;
}

/*
 * m_kick
 *
 * parv[0] = kickeur fou
 * parv[1] = chan
 * parv[2] = num victime
 * parv[3] = raison
 */
void m_kick(int sock, int parc, char **parv)
{
  aNick *k, *v;
  aChan *ch;
  aJoin *j = NULL;
  anAccess *a1 = NULL, *a2 = NULL;

  if(parc < 3) return;
  if(!(ch=getchaninfo(parv[1]))) return;
  if(!strcmp(cs.num,parv[2])) /*on sait jamais*/
  {
	  csjoin(ch);
	  csmode(ch, "+o %s", cs.num);
	  return;
  }

  if(!(v=num2nickinfo(parv[2]))) return;
  del_join(v, parv[1]);

  if(!v->user) return;/*v = pas auth*/
#ifdef N2K
  k = num2nickinfo(parv[0]);
#else
  k = getnickbynick(parv[0]);
#endif
  if(!k) return;
  if(IsAnAdmin(v->user))/*cas v = admin*/
  {
	  if(k->user && k->user->level > v->user->level) return; /*k = admin superieur*/
	  csmode(ch, "-o %s", k->numeric);/*k non auth, user ou admin inferieur*/
	  if((j = getjoininfo(k, ch->chan))) j->is_op = 0;
	  return;
  }
  else
  {
	  if(k->user && IsAnAdmin(k->user)) return;/*k=admin et v non*/
	  a1=GetAccessIbyUserI(v->user,parv[1]);
	  if(!a1 || !AProtect(a1) || ASuspend(a1)) return;/*v a pas dacces ou non protect*/
	  if(k->user)
	  {
		  a2=GetAccessIbyUserI(k->user,parv[1]);
		  if(a2 && a2->level >= a1->level && !ASuspend(a2)) return; /*meme niveau*/
		  csmode(a1->c, "-o %s", k->numeric);
		  if((j = getjoininfo(k, ch->chan))) j->is_op = 0;
		  return;
	  }
	  csmode(a1->c, "-o %s", k->numeric);
	  if((j = getjoininfo(k, ch->chan))) j->is_op = 0;
  }
  return;
}

/*
 * m_join
 *
 * parv[0] = nick
 * parv[1] = chans
 */
void m_join(int sock, int parc, char **parv)
{
  char *p, *ch = NULL, *ch2=NULL;
  aNick *nick;
  int i = 1;

  if(parc < 2) return;
#ifdef N2K
  nick = num2nickinfo(parv[0]);
#else
  nick = getnickbynick(parv[0]);
#endif
  if(strchr(nick->nick, '@')) return;
  if(parv[1][0]=='0') /* partall !!*/
  {
	  aJoin *join, *jt;
	  for(join=nick->joinhead;join;join=jt)
	  {
		  jt=join->next;
		  del_join(nick, join->channel);
	  }
	  return;
  }
  ch = strchr(parv[1], ',');
  if(ch && *ch==',') ch++;
  p = strtok(parv[1], ",");
  while(p != NULL)
  {
	  /*Toutes ces changes peuvent parraitres inutiles: mais sinon en cas de join #1,2,3 a foire (mem)*/
	  add_join(nick, p, 0, 0, time(NULL), 0);
	  i++;
	  if(!ch) break;
	  ch2 = strchr(ch, ',');
	  if(ch2 && *ch2==',') ch2++;
	  p = strtok(ch, ",");
	  ch = ch2;
  }

  return;
}

/*
 * m_create
 *
 * parv[0] = num
 * parv[1] = chan
 * parv[2] = timestamp
 */
void m_create(int sock, int parc, char **parv)
{
  char *p;
  aNick *nick=NULL;
  int i = 1;

  if(parc < 3) return;
  nick = num2nickinfo(parv[0]);
  p = strtok(parv[1], ",");
  while(p != NULL)
  {
	  add_join(nick, p, 1, 0, atol(parv[2]), 0);
	  i++;
	  p = strtok(NULL, ",");
  }
  return;
}

/*
 * m_setident
 *
 * parv[0] = num
 * parv[1] = host
 */
void m_setident(int sock, int parc, char **parv)
{
	aNick *who;

	if(parc<2) return;
	who=num2nickinfo(parv[0]);
	strncpy(who->ident, parv[1], USERLEN);
	return;
}

/*
 * m_sethost
 *
 * parv[0] = num
 * parv[1] = host
 */
void m_sethost(int sock, int parc, char **parv)
{
	aNick *who;

	if(parc<2) return;
	who = num2nickinfo(parv[0]);

	if(strchr(parv[1], ':') || !strchr(parv[1], '.') || parv[1][0]=='\0') return;

#ifdef HAVE_CRYPTHOST
	strncpy(who->crypt, parv[1], HOSTLEN);
#else
	strncpy(who->host, parv[1], HOSTLEN);
#endif

	return;
}

/*
 *m_nick
 *
 * parv[1] = nick
 * parv[2] = hop
 * parv[3] = timestamp
 * parv[4] = username
 * parv[5] = host
 *  ==[ user a un/des mode(s) | user a pas de mode ]==
 * parv[6] = modes | base64
 * parv[7] = base64 | num
 * parv[8] = num | name
 * parv[9] = name |
 */
void m_nick(int sock, int parc, char **parv)
{
	char *server = "", *ident, *nick, *host, *num=NULL, *lastnick = NULL, *ac=NULL, *crypt=NULL;
	aNick *who = NULL;
	anUser *user;
	struct Link *serv;
	int i=6;

	if(parc > 7)
	{  /*Nouveau nick*/
		ident = parv[4];
		if(*ident == '~') ident++;
		host = parv[5];
		if(*host=='+') return;
		nick = parv[1];

#ifdef HAVE_CRYPTHOST
		if(strchr(host, '.')) /* la clef est la premire partie de l'host*/
		{
			char key[32];
			key[0]='\0';
			if(!is_ip(host))
			{
				strncpy(key, host, 32);
				strtok(key, ".");
			}
			else strncpy(key, strchr(host, '.'), 32);
			crypt = hostprot(key, host); /* cryptage de l'host en md2 ou md5 suivant certains critres*/
		}
#endif
		if(parv[6][0] == '+')
		{
			i=7;
			if(strchr(parv[6], 'r')) ac=parv[i++];
		}

		if((serv=num2servinfo(parv[i+1]))) server = serv->serv;

#ifdef HAVE_CRYPTHOST
		who = add_nickinfo(nick, ident, host, crypt, parv[i], parv[i+1], server, parv[i+2], atol(parv[3]), (parv[6][0] == '+') ? parv[6] : "");
		free(crypt);
#else
		who = add_nickinfo(nick, ident, host, parv[i], parv[i+1], server, parv[i+2], atol(parv[3]), (parv[6][0] == '+') ? parv[6] : "");
#endif
		if(ac)
		{
			if(!(user=getuserinfo(ac)))
			{
				cswallops("ACCOUNT inconu (%s) pour %s depuis %s", ac, who->nick, server);
				cs_account(who, NULL);
				return;
			}
			if(user->n)
			{
				cswallops("ACCOUNT DESYNCH (%s) pour %s depuis %s", ac, who->nick, server);
				cs_account(who, NULL);
				return;
			}
			who->user=user;
			who->user->n=who;
			who->user->lastseend = time(NULL);
			csntc(who, "Vous avez t \002ridentifi\002");
			if(UProtect(user)) del_killinfo(who->nick, who->numeric); /* arret du kill */
		}
#ifdef USE_WELCOMESERV
		if((ConfFlag & CF_WELCOME) && !burst) /* Envoi du welcome si il est actif et non nul*/
			choose_welcome(who->numeric);
#endif
	}
	else if(parc <=3)
	{  /* Changement de pseudo */
		nick = parv[1];
		num = parv[0];
		who = num2nickinfo(num);

		del_killinfo(who->nick, num);
		lastnick = who->nick;
		strcpy(who->nick, nick);
	}
	else return;

	if(!who) who = num2nickinfo(num);/*on refait pas la loop*/

	if(IsBadNick(nick))
	{
		csntc(who, "\002Avertissement\002 : Vous venez de prendre un pseudo illgal");
		if(kill_interval)
			csntc(who, "Veuillez changer de pseudo sinon vous serez dconnect dans %d secondes.", kill_interval);
		add_killinfo(nick, num, time(NULL), 1, (ConfFlag & CF_CHANGENICK));
		return;/*on check meme pas le reste..*/
	}
	if(who->user && !strcasecmp(who->user->nick, nick)) return;/*le mec est auth sur ce nick > exit*/
	if(!(user = getuserinfo(nick))) return;/*le pseudo n'est pas reg, on evite leisregnick > exit*/
	csntc(who, "\002Avertissement\002 : Le pseudo %s est enregistr, si c'est le votre, identifiez vous via: /%s LOGIN %s <votre pass>", nick, cs.nick, nick);
	if(UProtect(user))
	{
		if(UKill(user)) csntc(who, "Veuillez changer de pseudo sinon vous serez dconnect dans %d secondes.", kill_interval);
		else csntc(who, "Veuillez changer de pseudo sinon votre pseudo sera automatiquement chang dans %d secondes.", kill_interval);
		add_killinfo(who->nick, who->numeric, time(NULL), 2, UKill(user));
	}
	else csntc(who, "Le propritaire peut le rcuprer en vous deconnectant.");
	return;
}

/*
 * m_burst
 *
 * parv[0] = num
 * parv[1] = chan
 * parv[2] = timestamp
 * parv[3] = modes / users
 * parv[4] = users / bans
 * parv[5] = bans
 */
void m_burst(int sock, int parc, char **parv)
{
  aNick *nick=NULL;
  aLink *j=NULL;
  aChan *c;
  char *modes = NULL, *chan = parv[1], *timestamp = parv[2], *users = NULL, *key = NULL, *limit = NULL;
  char *flags = NULL, *num = NULL, *pv = NULL/*, *bans = NULL*/;
  int op = 0, voiced = 0, i=4;
  long il = atol(timestamp);

  if(parc > 2 && parv[3][0] == '%') return;

  if(parc > 4 && parv[3][0] == '+')
  {  /* channel has mode(s) */
	modes = parv[3];
	if(strchr(modes, 'l')) { limit = parv[i]; i++; }
	if(strchr(modes, 'k')) { key = parv[i]; i++; }
	users = parv[i];
	/*if(parc > i) bans = parv[++i];*/
  }
  else
  {  /* channel has no mode */
	modes = "";
	users = parv[3];
	/*if(parc > 4) bans = parv[4];*/
  }

  if(!burst)
  {
	  if((c=getchaninfo(chan)))
	  {
		  if (!c->chanstamp || il < c->chanstamp)
		  { /* les users du chan vont se faire deop et les modes rinitialiss comme le veut le serveur qui envoi le burst*/
		      strncpy(c->modes, modes, 20);
		      if(limit)
		      {
				  strcat(c->modes, " ");
				  strncat(c->modes, limit, 10);
			  }
			  if(key) strncpy(c->key, key, 100);
			  putserv("%s %s %s +o %s\n", bot.servnum, TOKEN_MODE, c->chan, cs.num);
			  c->chanstamp = il;
			  for(j=c->members;j;j=j->next)
			  {
			  	/* le pb ici c'est qu'on ne peut dterminer un user qui s'est fait deop UNIQUEMENT sur les
				  chans regs, donc si un mec tait op sur un chan non reg, que le timestamp le fasse se deop,
				  et qu'il veut reg le chan, il pourra sans etre op (bien qu'il l'tait tout de mme avant)
				  Toute faon  part a on s'en fou vu que les chans sont pas reg on en a rien  battre */
			  	j->value.j->is_op=0;
				j->value.j->is_voice=0;
			  }
		  }
	  }
  }
  else
  {
	  if(!strcasecmp(cs.chan, chan))
	  {
		  putserv("%s %s %s %ld +mints %s:o,%s:o\n", bot.servnum, TOKEN_BURST, cs.chan, il, ws.num, cs.num);
		  burst=-1*burst; /* pour remplacer le 'gotcs' local, sans rajouter de global var*/
	  }
	  else if((c=getchaninfo(chan)) && c->suspend[0] == '\0')
	  {
	  	int gna=0, j=0;
		  putserv("%s %s %s %ld %s %s:o\n", bot.servnum, TOKEN_BURST, c->chan, il, c->defmodes, cs.num);
		  strncpy(c->modes, c->defmodes, 20);
		  strtok(c->modes, " ");
		  while(modes && *modes)
		  {
		  	gna=1;
			j=0;
		  	while(c->modes[j] != '\0')
			{
				if(c->modes[j] == *modes || *modes=='+')
				{
					gna=0;
					break;
				}
				j++;
			}
			if(gna) sprintf(c->modes, "%s%c", c->modes, *modes);
		  	modes++;
		  }
		  if(limit) { strcat(c->modes, " "); strcat(c->modes, limit); }
		  if(key) strcpy(c->key, key);
		  if(CLockTopic(c))
		  {
			  if(c->topic[0] != '\0') cstopic(c, "%s", c->topic);
			  else if(c->deftopic[0] != '\0') cstopic(c, "%s", c->deftopic);
		  }
		  SetCJoined(c);
		  c->chanstamp = il;
	  }
  }

  pv=users;
  while((num=strtok(pv, ",")))
  {
	if((flags = strchr(num, ':')))
	{
		op=0;
		voiced=0;
		*flags = '\0';
		while (*(++flags))
			if (*flags == 'o')
			  op = 1;
			else if (*flags == 'v')
			  voiced = 1;
			else
			  break;
	}
	add_join((nick=num2nickinfo(num)), chan, op, voiced, il, 1);
	pv=NULL;
  }
 /* if(bans)
  {
  	  bans++;
	  bans++;
	  pv=bans;
	  while((mask=strtok(pv, " ")))
	  {
	        add_ban(chan, mask, "Aucune", "MODE", time(NULL), 0, 0);
	  	pv=NULL;
	  }
  }*/
  return;
}

/*
 * m_quit
 *
 * parv[0] = nick
 * parv[1] = msg
 */
void m_quit(int sock, int parc, char **parv)
{
	aNick *nick;

#ifdef N2K
  nick = num2nickinfo(parv[0]);
#else
  nick = getnickbynick(parv[0]);
#endif
	if(!nick) return; /* !!!*/
	if(nick->user)
	{
#ifdef HAVE_TRACK
		struct track *track = NULL, *tt = NULL;
		extern struct track *trackhead;

		for(track=trackhead;track;track=tt)
		{
			tt = track->next;
			if(nick->user == track->tracked)
			{
				csntc(track->tracker, "[\2Track\2] %s (User: %s) vient de quitter.", nick->nick, track->tracked->nick);
				break;
			}
			if(nick == track->tracker) del_track(track->tracked);
		}
#endif
		nick->user->n=NULL;
	}
	del_killinfo(nick->nick, nick->numeric);
	del_nickinfo(nick);
	return;
}

/*
* m_kill
* Nettoyage des vars
*/

void m_kill(int sock, int parc, char **parv)
{
	char *num = parv[1];
	aNick *nick=NULL;

	if(!(nick = num2nickinfo(num))) return;
	if(nick->user)
	{
#ifdef HAVE_TRACK
		struct track *track = NULL, *tt = NULL;
		extern struct track *trackhead;

		for(track=trackhead;track;track=tt)
		{
			tt = track->next;
			if(nick->user == track->tracked)
			{
				csntc(track->tracker, "[\2Track\2] %s (User: %s) vient d'tre Killer.", nick->nick, track->tracked->nick);
				break;
			}
			if(nick == track->tracker) del_track(track->tracked);
		}
#endif
		nick->user->n=NULL;
	}
	del_nickinfo(nick);
	if(IsRegNick(nick->nick)) del_killinfo(nick->nick, num);
	return;
}

/*
 * m_away
 *
 * parv[0] = who
 * parv[1-parc] = raison
 */
void m_away(int sock, int parc, char **parv)
{
	aNick *nick;
#ifdef USE_MEMOSERV
	aMemo *memo;
	int i = 0;
#endif

#ifdef N2K
    if(!(nick = num2nickinfo(parv[0]))) return;
#else
    if(!(nick = getnickbynick(parv[0]))) return;
#endif
	if(parc>1) nick->away = 1;
	else
	{
		nick->away = 0;
#ifdef USE_MEMOSERV
		if(nick->user)
		{
			for(memo = nick->user->memohead;memo;memo=memo->next) i++;
			if(i) csntc(nick, "Vous avez reu \002%d\002 memos, pour les lire tapez \002/%s memos LIRE <ALL|N du memo>\002", i, cs.nick);
		}
#endif
	}
	return;
}

/*
 * m_whois
 *
 * parv[0] = who
 * parv[1] = ma num
 * parv[2] = pseudo du robot
 */
void m_whois(int sock, int parc, char **parv)
{
	char insulte[200];
	int i;
	aNick *tmp;

	insulte[0] = '\0';
	if(!strcasecmp(parv[2], cs.nick))
		putserv("%s 311 %s %s %s %s * :%s\n", bot.servnum, parv[0], cs.nick, cs.ident, cs.host, cs.name);
#ifdef USE_WELCOMESERV
	else if(!strcasecmp(parv[2], ws.nick))
		putserv("%s 311 %s %s %s %s * :%s\n", bot.servnum, parv[0], ws.nick, ws.ident, ws.host, ws.name);
#endif
	else
	{
	  tmp = getnickbynick(parv[2]);
	  if(!tmp) return;

	  i = rand()%10;
	  if(i == 0) strcpy(insulte, "a une face de cul de babouin !");
	  else if(i == 1) strcpy(insulte, "est une lamouze !");
	  else if(i == 2) strcpy(insulte, "est transsexuel !");
	  else if(i == 3) strcpy(insulte, "a un QI ngatif !");
	  else if(i == 4) strcpy(insulte, "a la diahrre !");
	  else strcpy(insulte, "est completement dbile !");
#ifdef HAVE_CRYPTHOST
	  putserv("%s 311 %s %s %s %s * :%s\n", bot.servnum, parv[0], parv[2], tmp->ident, tmp->crypt, tmp->name);
#else
	  putserv("%s 311 %s %s %s %s * :%s\n", bot.servnum, parv[0], parv[2], tmp->ident, tmp->host, tmp->name);
#endif
   	  putserv("%s 312 %s %s %s :%s\n", bot.servnum, parv[0], parv[2], tmp->server, tmp->name);
	  putserv("%s 390 %s %s :%s\n", bot.servnum, parv[0], parv[2], insulte);
	  if(tmp->is_oper) putserv("%s 313 %s %s :est un IRC Oprateur tyrannique et pervers\n", bot.servnum, parv[0], parv[2]);
	  putserv("%s 317 %s %s 99999999 99999999 :Secondes d'idle, secondes d'uptime\n", bot.servnum, parv[0], parv[2]);
	  putserv("%s 318 %s %s :Fin de la commande /WHOIS\n", bot.servnum, parv[0], parv[2]);
	  return;
  	}
	putserv("%s 312 %s %s %s :%s\n", bot.servnum, parv[0], parv[2], bot.server, bot.name);
	putserv("%s 390 %s %s :est un Robot\n", bot.servnum, parv[0], parv[2]);
	putserv("%s 313 %s %s :est un IRC Oprateur - Service\n", bot.servnum, parv[0], parv[2]);
	putserv("%s 317 %s %s %ld %ld :Secondes d'idle, secondes d'uptime\n", bot.servnum, parv[0], parv[2], time(NULL) - tidle, tuptime);
	putserv("%s 318 %s %s :Fin de la commande /WHOIS\n", bot.servnum, parv[0], parv[2]);
	return;
}

void m_topic(int sock, int parc, char **parv)
{
	aChan *chan = NULL;
	char *c = parv[1];
	if(!(chan = getchaninfo(c))) return;
	if(CLockTopic(chan))
	{
		if(chan->topic[0] != '\0') cstopic(chan, "%s", chan->topic);
		else cstopic(chan, "%s", chan->deftopic);
	}
#ifdef N2K
	else strncpy(chan->topic, parv[3], TOPICLEN);
#else
	else strncpy(chan->topic, parv[2], TOPICLEN);
#endif
	return;
}

/*
 * m_eob
 *
 * parv[0] = who
 * parv[1] = me :)
 */
void m_eob(int sock, int parc, char **parv)
{
	aChan *chan;
	struct timeval tv;
	struct Link *serv=NULL;
	float tt;

	if (!(serv=num2servinfo(parv[0])))
	{
		putlog(LOG_ERREURS, "ATTENTION: Prfixe serveur reu incorrect\n");
		return ;
	}
	if(!(serv->flag & ST_BURST))
	{
		putlog(LOG_ERREURS, "ATTENTION: Erreur interne  m_eob()\n");
		return ;
	}
	serv->flag &= ~ST_BURST;
	serv->flag |= ST_ONLINE;
	if(serv == mainhub)
	{
		if(burst>0) putserv("%s %s %s %ld +mints %s:o,%s:o\n", bot.servnum, TOKEN_BURST, cs.chan, time(NULL), cs.num, ws.num);
		else if(burst < 0) burst = -1*burst;

		for(chan = chanhead;chan;chan=chan->next)
		{
			if(!CJoined(chan) && chan->suspend[0] == '\0')
			{
				putserv("%s %s %s %ld %s %s:o\n", bot.servnum, TOKEN_BURST, chan->chan, time(NULL), chan->defmodes, cs.num);
				if(chan->topic[0] != '\0') cstopic(chan, "%s", chan->topic);
				else if(chan->deftopic[0] != '\0') cstopic(chan, "%s", chan->deftopic);
				SetCJoined(chan);
				chan->chanstamp = time(NULL);
			}
		}

		putserv("%s %s\n", bot.servnum, TOKEN_EOB);
		putserv("%s %s\n", bot.servnum, TOKEN_EOBACK);

		if(dataQ)
		{
			gettimeofday(&tv, NULL);/*un peu de stats sur le burst pour *fun* */
			tt = (tv.tv_usec - burst)/1000000.0;
			if(tt<0) tt= -1*tt;
			cswallops("Synchronisation de %d bytes en %12.6f s (Taux %12.6f Ko/s)", dataQ, tt,(float) (dataQ / 1024.0) / tt);
		}
		burst=0;
	}
	alarm(ALARMINTERVAL);
	return;
}

void ClearNicksbyServ(char *serv)
{
	aNick *nick, *nt;

	for(nick = nickhead;nick;nick=nt)
	{
		nt=nick->next;
		if(!strcasecmp(nick->server, serv))
		{
			if(nick->user) nick->user->n = NULL;
			del_nickinfo(nick);
		}
	}
}

void m_server(int sock, int parc, char **parv)
{
  struct Link *link, *last;

  link=calloc(1,sizeof(struct Link));
  strncpy(link->serv,parv[1],200);
#ifdef N2K
  strncpy(link->num,parv[6],2);
#else
  strncpy(link->num,parv[6],1);
#endif
  link->flag=0;
  link->next=NULL;
  if(atoi(parv[2])==1)/*1erhub*/
  {
	  link->hub = NULL;
	  link->last = NULL;
	  mainhub=link;
	  registered = 1;
  }
  else
  {
	  last=num2servinfo(parv[0]);
	  link->hub = last; /* ptr vers son Hub */
	  last->flag |= ISHUB;
	  for(last=mainhub;last->next;last=last->next); /* get last ptr*/
	  last->next = link; /* chainage positif */
	  link->last = last; /* chainage ngatif */
  }

  if (parv[5][0] == 'J') link->flag |= ST_BURST;
  else link->flag |= ST_ONLINE;
  return;
}

void NullHub(struct Link *hub)
{
	struct Link *l;
	for(l=hub;l;l=l->next) if(l->hub == hub) l->hub = NULL;
}

void m_squit(int sock, int parc, char **parv)
{
	struct Link *link=NULL, *last=NULL, *lt=NULL;

	/*nettoyage des serveurs en "dessous" (dont il est Hub) */

	for(link=mainhub;link && strcasecmp(parv[1], link->serv);link=link->next);
	if(link)
	{
		if(link == mainhub) /* squit de l'uplink, tentative de reconnexion.. */
		{
			for(;link;link=lt)
			{
				lt=link->next;
				free(link);
			}
			mainhub = NULL;
			return;
		}

		if(link->next)
		{
			link->next->last = link->last;
			link->last->next = link->next;
		}
		else link->last->next = NULL;

		if(link->flag & ISHUB)
		{
			NullHub(link); /* s'il est hub ya forcement un leaf (next) apres!? */
			for(last=link->next;last;last=lt)
			{
				lt=last->next;
				if(!last->hub)
				{
					ClearNicksbyServ(last->serv);
					if(last->flag & ISHUB) NullHub(last);
					last->next->last = last->last;
					last->last->next = last->next;
					free(last);
					last = NULL;
				}
			}
		}
		ClearNicksbyServ(link->serv);
		free(link);
		link=NULL;
	}

	return;
}

void m_pass(int sock, int parc, char **parv)
{
	struct timeval tv;

	gettimeofday(&tv, NULL);
	burst = tv.tv_usec;
	putlog(LOG_PARSES, "Connect au serveur %s\n", bot.ip);
	registered = 0;
	if(!tuptime) tuptime = time(NULL);
	if(!lastresetlogs) lastresetlogs = time(NULL);
	if(!lastresetparse) lastresetparse = time(NULL);
	tidle = time(NULL);
	return;
}

void m_ping(int sock, int parc, char **parv)
{
    putserv("%s %s %s\n", bot.servnum, TOKEN_PONG, parv[1]);
    return;
}

void exec_cmd(unsigned int i, unsigned int securise, char *ni, unsigned int parc, char **parv)
{
	char tmp3[400], buff[512];
	aChan *chaninfo = NULL;
	anAccess *access = NULL;
	aNick *nick = NULL;

#ifdef N2K
	if(!(nick=num2nickinfo(ni))) return; /* !!!!*/
#else
	if(!(nick=getnickbynick(ni))) return; /* !!!!*/
#endif

	parv2msgn(parc, parv, 1, tmp3, 400);
	if(cs_cmds[i].secure && !securise && (strcasecmp(cs_cmds[i].realcmd, "memos") ||
	!strcasecmp(parv[1], "SEND")))
	{
		csntc(nick, "La commande \002%s\002 demande des informations prives. Veuillez utiliser \002/%s %s %s\002.", cs_cmds[i].cmd, cs.nick, cs_cmds[i].cmd, tmp3);
		return;
	}

	cs_cmds[i].used++;

	if(parc < cs_cmds[i].args)
	{
		syntax_cmd(nick, i);
		return;
	}
	if(isignore(nick) || checkflood(nick)) return;
	if(parc && cs_cmds[i].chan)
	{
		chaninfo = getchaninfo(parv[1]);
		if(nick->user) access = GetAccessIbyUserI(nick->user, parv[1]);
	}
	if(cs_cmds[i].level)
	{
		if(!nick->user || !nick->user->n) {
			csntc(nick, "%s : Vous n'tes pas logu ou vous n'avez pas encore enregistr d'Username", pasdeperm); return; }
		else if(cs_cmds[i].level < 5 && nick->user->level < cs_cmds[i].level) {
			csntc(nick, "%s : Vous devez tre admin pour utiliser cette commande", pasdeperm); return; }
		else if(cs_cmds[i].level >= 5 && parc && parv[1][0]=='#' && !IsAnAdmin(nick->user))
		{
			if(!access) {
				csntc(nick, "%s : Vous n'avez pas accs sur le salon %s", pasdeperm, parv[1]); return; }
			else if((access && cs_cmds[i].level > access->level)) {
				csntc(nick, "%s : Votre niveau sur %s est trop petit (Vous: \002%d\002, Commande: \002%d\002)", pasdeperm, parv[1], access->level, cs_cmds[i].level); return; }
			else if(access && ASuspend(access)) { csntc(nick, "%s : Votre accs est suspendu sur %s", pasdeperm, parv[1]); return; }
		}
	}
	if(cs_cmds[i].disable && strcasecmp(cs_cmds[i].realcmd, "DISABLE")) csntc(nick, "\002La commande %s est dsactive.", cs_cmds[i].cmd);
	else if(cs_cmds[i].chan && !chaninfo && strcasecmp(cs_cmds[i].realcmd, "REGCHAN")) csntc(nick, NOSUCHAN, parv[1]);
	else if(cs_cmds[i].chan && chaninfo->suspend[0] != '\0') csntc(nick, "%s est suspendu : %s", parv[1], chaninfo->suspend);
	else
	{
		char tmp[2 * NICKLEN];
#ifdef HAVE_TRACK
		struct track *track = NULL;
#endif
		snprintf(tmp, 2 * NICKLEN, "%s%s%s", nick->nick, nick->user ? "@" : "", nick->user ? nick->user->nick : "");
		if(cs_cmds[i].secure == 1 && parc) sprintf(buff, "\002%s\002 %s par %s", cs_cmds[i].cmd, parv[1], tmp);
		else if(cs_cmds[i].secure == 2 && parc) sprintf(buff, "\002%s\002 par %s", cs_cmds[i].cmd, tmp);
		else if(parc >= 2
			&& ((!strcasecmp(cs_cmds[i].realcmd, "USER") && !strcasecmp(parv[1], "NEWPASS"))
			|| (!strcasecmp(cs_cmds[i].realcmd, "SET") && !strcasecmp(parv[1], "PASS"))))
			sprintf(buff, "\002%s\002 %s par %s", cs_cmds[i].cmd, parv[1], tmp);
		else if(parc) sprintf(buff, "\002%s\002 %s par %s", cs_cmds[i].cmd, tmp3, tmp);
		else sprintf(buff, "\002%s\002 par %s", cs_cmds[i].cmd, tmp);

#ifdef HAVE_TRACK
		if(nick->user && (track=istrack(nick->user))) csntc(track->tracker, "[\2Track\2] %s", buff);
#endif
		putchan(buff);
		putlog(LOG_CMDS, "%s\n", buff);
		cs_cmds[i].func(nick, chaninfo, parc, parv);/*permet de passer la struct chaninfo (economie de loop)*/
	}
	return;
}

/*
 * m_privmsg
 *
 * parv[0] = envoyeur
 * parv[1] = destinataire (pseudo du robot, set dans botnick)
 * parv[2] = message reu.
 *
 * pparv[0] = robot qui a reu le message
 * pparv[pparc] = arguments de la commande
 * ATTENTION: pparc est le nb d'arg  partir de *parv[1]*
 */
void m_privmsg(int sock, int parc, char **parv)
{
	char *tmp, *cmd = NULL, *botnick = NULL;
	char *pparv[500], tmp2[500];
	aNick *ni=NULL;
	unsigned int i = 0, pparc = 0, securise = 0;

	tmp2[0]='\0';

	if(*parv[1] != '#')
	{  /* Message envoy  un robot*/

		pparv[1] = NULL;

		if(strchr(parv[1], '@'))
		{
			botnick = strtok(parv[1], "@");
			securise = 1;
		}
		else botnick = parv[1];

		if(parv[2][0]=='\0' || parv[2][0]==' ') return;
		tmp = strtok(parv[2], " ");
		if (tmp != NULL)
		{
			cmd = tmp;
			pparv[0] = cs.nick;
			tmp = strtok(NULL, " ");
		}
		while (tmp != NULL)
		{
			pparc++;
			pparv[pparc] = tmp;
			tmp = strtok(NULL, " ");
		}
#ifdef N2K
		if(strcasecmp("HELLO", cmd) && premiere && (ni = num2nickinfo(parv[0])))
#else
		if(strcasecmp("HELLO", cmd) && premiere && (ni = getnickbynick(parv[0])))
#endif
		{
			csntc(ni, "%s n'a pas encore t enregistr", cs.nick);
			return;
		}
		if (!strcmp(botnick, cs.num) || !strcasecmp(botnick, cs.nick))
		{
			for (i=0; cs_cmds[i].cmd[0] != '\0'; i++)
				if (((char *)strcasecmp(cs_cmds[i].cmd,cmd))==NULL)
				{
					exec_cmd(i, securise, parv[0], pparc, pparv);
					return;
				}
#ifdef N2K
			if(*cmd != '\001' && (ni = num2nickinfo(parv[0]))) csntc(ni, "La commande \002%s\002 n'existe pas.", cmd);
#else
			if(*cmd != '\001' && (ni = getnickbynick(parv[0]))) csntc(ni, "La commande \002%s\002 n'existe pas.", cmd);
#endif
		}/*endif*/
		return;
	}
	else
	{  /* Message envoy sur un salon (pour chanserv)*/
		if(parv[2][0] == bot.cara && parv[2][1] != '\0')
		{
			parv[2]++;

			if(parv[2][0]=='\0' || parv[2] == NULL) return;

			tmp = strtok(parv[2], " ");
			if (tmp != NULL)
			{
				cmd = tmp;
				pparv[0] = cs.nick;
				tmp = strtok(NULL, " ");
			}

#ifdef N2K
			if(strcasecmp("HELLO", cmd) && premiere && (ni = num2nickinfo(parv[0])))
#else
			if(strcasecmp("HELLO", cmd) && premiere && (ni = getnickbynick(parv[0])))
#endif
			{
				csntc(ni, "%s n'a pas encore t enregistr", cs.nick);
				return;
			}

			for (i=0; cs_cmds[i].cmd[0] != '\0'; i++)
				if (((char *)strcasecmp(cs_cmds[i].cmd,cmd))==NULL)
				{
					do
					{
						if(cs_cmds[i].chan && pparc == 0 && (tmp == NULL || tmp[0] != '#'))
						{
							pparc++;
							pparv[pparc] = parv[1];
							strcat(tmp2, " ");
							strcat(tmp2, parv[1]);
						}
						else if (tmp != NULL)
						{
							pparc++;
							pparv[pparc] = tmp;
							strcat(tmp2, " ");
							strcat(tmp2, tmp);
							tmp = strtok(NULL, " ");
						}
					}
					while (tmp);
					exec_cmd(i, securise, parv[0], pparc, pparv);
					return;
				}
		}
	}
	return;
}

void parse_this(int sock, char *msg)
{
	int parc = 0, i;
	char *ptr, *ptr2, *cmd;
	char *parv[MAXPARA + 2 ]; /* leave room for prefix and end-of-params NULL */
#ifdef DEBUG
	char logg[512];
	strncpy(logg,msg,512);
#endif

	dataQ += strlen(msg);/*comptage des bytes lues*/

	i=strlen(msg)-1;
	if (msg[i] == '\n') msg[i] = 0;
	i=strlen(msg)-1;
	if (msg[i] == '\r') msg[i] = 0;

	ptr = strchr(msg, ' ');
	if (!ptr) return;


#ifndef  N2K
	if (msg[0] == ':')
	{
		/* long prefix */
		parv[0] = &msg[1];
		*ptr = '\0';
		ptr++;
	}
	else if (ptr-msg && strncmp(msg, "SERVER", 6) && strncmp(msg, "PASS",4) && (registered))
#else
	if (ptr-msg && strncmp(msg, "SERVER", 6) && strncmp(msg, "PASS",4) && (registered))
#endif
	{
		/* numeric prefix */
		parv[0] = msg;
		*ptr = '\0';
		ptr++;
	} else
	{
		/* no prefix */
		parv[0] = bot.server;
		ptr=msg;
	}

	ptr2 = strchr(ptr, ' ');
	cmd = ptr;
	if (ptr2)
	{
		*ptr2 ='\0';
		ptr2++;
		ptr = ptr2;
		for (parc=1; (parc <= MAXPARA) && ptr; parc++)
		{
			ptr2 = strchr(ptr, ' ');
			if (!ptr2)
			{
				if (*ptr == ':') ptr++;
				parv[parc]=ptr;
				break;
			} else
			{
				if (*ptr == ':')
				{
					ptr++;
					parv[parc] = ptr;
					break;
				}
				*ptr2='\0';
				ptr2++;
				if (*ptr == ':')
				{
					ptr++;
					parv[parc] = ptr;
					break;
				}
				parv[parc] = ptr;
				ptr = ptr2;
			}
		}/*endfor*/

	} else parc=0;
	parc++;
	parv[parc] = NULL;

#ifdef DEBUG /*etical loging*/
	if(!(parc > 1 && ((!strcasecmp(cmd,TOKEN_PRIVMSG) && parv[1][0] == '#') ||
         !strcasecmp(cmd, TOKEN_PING) || !strcasecmp(cmd, TOKEN_PONG))))
	putlog(LOG_PARSES, "R - %s", logg);
#endif

	for (i=0; msgtab[i].cmd != NULL; i++)
	{
		if (!strcasecmp(msgtab[i].cmd,cmd))
		{
			msgtab[i].func(sock, parc, parv);
			return;/*on sait jamais */
		}
	}
}

int init_bot(int ip, int port)
{
  int sock, r, LTS;
  struct sockaddr_in fsocket;

  LTS = time(NULL);

  sock = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
  if (sock < 0) return(-1);

  fsocket.sin_family = AF_INET;
  fsocket.sin_addr.s_addr = ip;
  fsocket.sin_port = htons(port);
  r=connect(sock, (struct sockaddr *) &fsocket, sizeof(fsocket));

  if (r==(-1))
  {
    close(sock);
    return(-1);
  }

  bot.sock = sock;
  putserv("PASS %s\n", bot.pass);
#ifdef N2K
  putserv("SERVER %s 1 %u %u J10 %s]]] :%s\n", bot.server, STS, LTS, bot.servnum, bot.name);
#else
  putserv("SERVER %s 1 %u %u J10 %s]] :%s\n", bot.server, STS, LTS, bot.servnum, bot.name);
#endif

#ifdef USE_CHANSERV
	putserv("%s %s %s 1 %u %s %s %s B]AAAB %s :%s\n", bot.servnum, TOKEN_NICK, cs.nick, LTS, cs.ident, cs.host, cs.mode, cs.num, cs.name);
#ifdef HAVE_CRYPTHOST
	add_nickinfo(cs.nick, cs.ident, cs.host, cs.host, "B]AAAB", cs.num, bot.server, cs.name, LTS, cs.mode);
#else
	add_nickinfo(cs.nick, cs.ident, cs.host, "B]AAAB", cs.num, bot.server, cs.name, LTS, cs.mode);
#endif
#endif

#ifdef USE_WELCOMESERV
	putserv("%s %s %s 1 %u %s %s +kpd B]AAAB %s :%s\n", bot.servnum, TOKEN_NICK, ws.nick, LTS, ws.ident, ws.host, ws.num, ws.name);
#ifdef HAVE_CRYPTHOST
	add_nickinfo(ws.nick, ws.ident, ws.host, cs.host, "B]AAAB", ws.num, bot.server, ws.name, LTS, "kpd");
#else
	add_nickinfo(ws.nick, ws.ident, ws.host, "B]AAAB", ws.num, bot.server, ws.name, LTS, "kpd");
#endif
#endif

  return sock;
}

void run_bot(int sock)
{
  int maxi, r, t =0;
  fd_set rfds;
  char msg[512], buf[1024], c;
  struct timeval tv;
  int next_ping;
  int waitping;

  if (sock < 0 ) return;

  next_ping = time(NULL) + PING_FREQ;
  waitping = 0;

  while (!want_to_quit)
  {

    tv.tv_sec = 1;
    tv.tv_usec = 0;
    FD_ZERO(&rfds);
    FD_SET(sock, &rfds);
    maxi=sock;

    if (!FD_ISSET(sock, &rfds))
    {
      if (next_ping <= time(NULL))
      {
        if (waitping == 0)
        {
          putserv("%s %s :%s\n", bot.servnum, TOKEN_PING, bot.server);
          next_ping = time(NULL) + PING_FREQ;
          waitping = 1;
        } else
        {
          printf("Pas de rponse pour %s\n", bot.ip);
          close(sock);
          return;
        }
      }
      continue;
    } else
    {
      r=recv(sock, &c, 1, 0);
      if (r <= 0) break;
      next_ping = time(NULL) + PING_FREQ;
      waitping = 0;

      buf[t] = c;
      t++;
      if (c=='\n')
      {
        buf[t] = '\0';
        strncpy(msg, buf,512);
        parse_this(sock, msg);
        t=0;
      }
    }
  }
  if (want_to_quit) close(sock);
  return;
}
